/*********************************************************************
 * libserial - A library of routines for talking on a serial port
 * Author:  D. T. Spayde
 * Time-stamp: <2003-02-20 16:44:45 spayde>
 ********************************************************************/

#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <termios.h>
#include <string.h>
#include <errno.h>
#include <time.h>

static const int kMaxReadLoop = 10;  /* Number of times to check
					serial port for input during a
					given read command */
static const char* kTerminator = "\r\n";      /* Serial terminator */
static const struct timespec kWriteWait = { 0, 50000000 }; /* Wait time
							      between
							      serial
							      write
							      commands
							      (50 msec)
							   */
static const struct timespec kReadWait = { 0, 10000000 }; /* Wait time
							     between
							     serial
							     port
							     checks
							     (10 msec)
							   */
static const int kMaxLength = 100;

//////////////////////////////////////////////////////////////////////
// Opens the serial port specified by sPort
// Return value:
//	>0 = file descriptor for the port
//      <0 = error
//////////////////////////////////////////////////////////////////////
int open_port(const char *sPort) {

    //    printf("**open_port => sPort = %s\n", sPort);

    // Open port and attach to file descriptor fd
    // Modes:
    //	O_RDWR = read/write
    //	O_NOCTTY = port not attached to terminal, no echoing
    //	O_NONBLOCK = non-blocking mode (return from read/write
    // immediately)
    int fd;
    fd = open(sPort, O_RDWR | O_NOCTTY | O_NONBLOCK);

    if (fd < 0) {
	// Failed to open serial port
	printf("**open_port => Open error %d %s\n", errno, strerror(errno));
    } else {
	// Succeeded in opening serial port
	// Do port configuration; discard current configuration
	struct termios new_termios;

	// Read current attributes into termios struct
	tcgetattr(fd, &new_termios);
	// Flush unread and unwritten data from buffer
        tcflush(fd, TCIFLUSH);
        tcflush(fd, TCOFLUSH);
        
	// Control Modes
	//	CS7:  7 bit characters
	//	CREAD:  Read characters
	//	CLOCAL:  No modem
	//	HUPCL:  Hang up
	//	PARENB:  Enable parity
	//	PARODD:  Set odd parity
        new_termios.c_cflag = CS7 | CREAD | CLOCAL | HUPCL| PARENB | PARODD;
	// Input Modes
	//	ICRNL:  translate CR to NL on input
	new_termios.c_iflag &= ~ICRNL;
	// Local Modes
	new_termios.c_lflag = 0;
        
	// Set baud rate of 9600 bps
        cfsetospeed(&new_termios, B9600);
	
	// Make changes take effect immediately
        tcsetattr(fd, TCSANOW, &new_termios);
    }
    //    printf("**open_port => fd = %d\n", fd);
    return fd;
} // End open_port

//////////////////////////////////////////////////////////////////////
// Writes psOutput on the serial port specified by fd
// Return value:
//	>= 0 => number of characters written
//	-1 => write failed
//////////////////////////////////////////////////////////////////////
int write_port(const int fd, char *psOut) {

    //    printf("**write_port => psOut = %s fd = %d\n", psOut, fd);
    int iOut;	// number of characters written to fd

    // Check availability of port
    if (fd < 1) {
	// Port not open (should never get here)
	printf("** write_port => Port not open\n");
	return -1;
    }

    // Append terminator
    char psOutput[strlen(psOut)+strlen(kTerminator)+1];
    strcpy(psOutput, psOut);
    strcat(psOutput, kTerminator);

    // Write to port
    iOut = write(fd, psOutput, strlen(psOutput));
    if (iOut < 0) {
	// Write failed
	printf("** write_port => Write failed \n");
    } else {
	// Write succeeded
	;
    }

    nanosleep(&kWriteWait, NULL);

    return iOut;

} // End write_port

//////////////////////////////////////////////////////////////////////
// Reads input from serial port specified by fd and stores result in
// psInput
// Return value:
//	>= 0 => number of characters read
//	-1 => read failed
//////////////////////////////////////////////////////////////////////
int read_port(const int fd, char* psInput, int iMax) {

    // Check availability of port
    if (fd < 1) {
	// Port not open
	printf("** read_port => Port not open\n");
	return -1;
    } 

    // Read in data from port
    char sIn[kMaxLength];  // Buffer to hold input data
    strncpy(sIn, "\0", kMaxLength);
    int iRead=0;	// number of characters read
    int iIn=0;		// total number of characters read
    int iCount=0;	// number of times through read loop
    //    iIn = read(fd, psInput, iMax-1);

    // Loop until collect termination string "\r\n"
    for (iCount = 0; iCount < kMaxReadLoop; iCount++) {
	// Get number of characters waiting to be read
	ioctl(fd, FIONREAD, (int)&iRead);

	// Read in all available characters
	if (iRead != 0 && kMaxLength > (iIn + iRead)) {
	    read(fd, &sIn[iIn], iRead);
	    iIn += iRead;
	}

	// Check to see if termination string has been detected
	// Remove termination string and append '\0'
	char* term_p = NULL;  /* pointer to termination string in psInput */
	//	printf("**read_port => iCount = %d sIn = %s\n", iCount, sIn);
	term_p = strstr(sIn, kTerminator);
	if (term_p != NULL) {
	    /* Found termination string */
	    /* Eliminate terminator by putting in NULL at pointer to
	     *beginning of terminator */
	    *term_p = '\0';
	    break;
	}
	    
	nanosleep(&kReadWait, NULL);

    } // End read loop

    if (iIn < 0) {
	// Read failed for some reason
	if (errno == EAGAIN) {
	    // Non-blocking mode error code
	    // No input immediately available
	    // Assume there was nothing to read
	    return 0;
	} else {
	    // All other failure modes
	    printf("** read_port => Read error %d %s\n", errno, 
		   strerror(errno));
	}
    } else {
	// Read succeeded

	int iIn = strlen(sIn);
	strncpy(psInput, sIn, iMax-1);

	// Serial device communicates in 7 bit characters so mask off
	// top most bit
	int i;
	for (i=0; i < iIn; i++) {
	    psInput[i] &= 0x7f;
	    //	    printf("psInput[%d] = 0x%2.2x %c \n", i, psInput[i], psInput[i]);
	}
    }

    //    printf("**read_port => %s\n", psInput);

    return iIn;

} // End read_port

//////////////////////////////////////////////////////////////////////
// Closes the serial port specified by fd
//////////////////////////////////////////////////////////////////////
void close_port(const int fd) {

    if (fd > 0) {
	close(fd);
    }

} // End close_port
