/****************************************************************
 *
 * smsTempServ
 * Author:  Damon Spayde (spayde@jlab.org)
 * Time-stamp: <2003-08-11 13:50:17 spayde>
 * 
 ***************************************************************/

/****************************************************************
 * 
 ***************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <unistd.h>
#include <ctype.h>
#include <time.h>
#include <errno.h>
#include <sys/stat.h>
#include "smsServCommon.h"
#include "libserial.h"

/****************************************************************
 *** Function Declarations
 ***************************************************************/
int monitor_initialization(void);
int monitor_poll(void);
int get_options(int argc, char* argv[]);

/****************************************************************
 *** Global Variable Declarations
 ***************************************************************/
char* ini_filename; /* Name of initalization file */
char* serial_port;  /* Serial port of monitor */
char* monitor_id;   /* Name of monitor in ini file */
char* pipe_filename;/* Name of output pipe */
unsigned int monitor_num; /* Number of monitor */
char poll_all_cmd[kMaxLength]; /* Command to query all channels */
int plc_addr[kNumChannels];    /* Array to hold PLC memory locations */
int fd_serial;      /* File descriptor of serial port */
FILE* out_file;     /* File handle for output file */

/****************************************************************
 ************************ Main Program **************************
 ***************************************************************/
int main(int argc, char *argv[]) {

    int ret = 0;

    /* 
     * Process the command line
     */
    ret = get_options(argc, argv);
    if (ret != 0) {
	fprintf(stderr, "Encountered error in get_options().  Exiting . . .\n");
	exit(1);
    }

    /*
     * Initialize the monitor
     */
    //    fprintf(stderr, "Calling monitor_initialization . . .\n");
    ret = monitor_initialization();
    //    fprintf(stderr, "monitor_initialization returned with value %d. . .\n", ret);
    if (ret != 0) {
	fprintf(stderr, "Encountered error in monitor_initialization().  Exiting . . .\n");
	exit(1);
    }

    /*
     * Poll the monitor indefinitely
     */
    while (1) {
	//	fprintf(stderr, "Calling monitor_poll . . .\n");
	//	usleep(500000);
	ret = monitor_poll(); 
	//	fprintf(stderr, "monitor_poll returned with value %d. . .\n", ret);
	if (ret != 0) {
	    fprintf(stderr, "Encountered error in monitor_poll().  Exiting . . .\n");
	    exit(1);
	}
    }
    close_port(fd_serial);
    fclose(out_file);

    return 0;
}


/***************************************************************/
/***************************************************************/
/*********************** Functions *****************************/
/***************************************************************/
/***************************************************************/

/****************************************************************
 * This function parses the command line and insures all necessary
 * information is there
 ***************************************************************/
int get_options(int argc, char* argv[]) {
    ini_filename = NULL;
    serial_port = NULL;
    monitor_id = NULL;
    pipe_filename = NULL;
    monitor_num = 0;
    int c;

    while ((c = getopt(argc, argv, "d:m:i:p:")) != -1) {
	switch (c) {
	case 'd':
	    /* Serial Port */
	    serial_port = optarg;
	    break;
	case 'm':
	    /* Monitor ID */
	    monitor_id = optarg;
	    /* Extract number of monitor from monitor_id */
	    int i = 0;
	    for (i = 0; i<strlen(monitor_id); i++) {
		if (isdigit(monitor_id[i])) {
		    monitor_num = atoi(&monitor_id[i]);
		    //		    printf("Monitor number is %d\n", monitor_num);
		    break;
		}
	    }
	    if (monitor_num <= 0) {
		fprintf(stderr, 
			"**monitor_initialization => Unable to determine monitor number.  Exiting . . .\n");
		return -1;
	    }
	    break;
	case 'i':
	    /* ini file name */
	    ini_filename = optarg;
	    break;
	case 'p':
	    /* Output pipe file name */
	    pipe_filename = optarg;
	    break;
	case '?':
	    if (isprint(optopt)) {
		fprintf(stderr, "Unknown option `-%c`.\n", optopt);
	    } else {
		fprintf(stderr, "Unknown option character `%x`.\n", optopt);
	    }
	    return 1;
	default:
	    return 1;
	}
    }

    /* Must specify all four on command line */
    if (serial_port == NULL || monitor_id == NULL 
	|| ini_filename == NULL || pipe_filename == NULL) {
	fprintf(stderr, "usage:  smsTempServ -d <serial port> -m <monitor id> -i <ini filename> -p <pipe filename>\n");
	return 1;
    }

    return 0;
}

/****************************************************************
 * This routine parses the file <name> containing the temperature
 * curves, monitor parameters, etc.
 * Return Value = file descriptor of serial port
 ****************************************************************
 * File Format:
 *---------------------------------------------------------------
 * Comment lines begin with ";"
 * Section headers enclosed in square brackets.  Also delimit
 * sections.
 * Three types of sections:  temperatures curves, monitors, and PLC
 * Temperature Curve Fields:
 *	SerialNo:  curve serial no, up to 10 characters
 *	Format:  curve data format (2=V/K, 3=Ohm/K, 4=log Ohm/K)
 *	LimitKelvin:  curve upper T limit in K
 *	Coefficient:  curve T coefficient (1=negative, 2=positive)
 *	PointCount:  number of points in curve (2-200)
 *	PointN:  curve point N, format is PointN = input units, T
 *		curve points not in any order, must be sorted before
 *		putting in monitor.
 * PLC Fields:
 *	HardwareAddress:  MAC address of the PLC
 *	MinCCMmSec:  minimum time between writes to PLC
 *	KeepAliveAddress:  V memory address of keep alive bit
 *	KeepAliveBitNumber:  specifies bit number used for keep alive
 *		(use 0 if not specified)
 *; NOTE: To toggle the bit, the entire 16-bit register must be read in,
 *;       the bit in question AND-ed out or OR-ed in, and then the whole
 *;       register written back out again.  It may be desirable to allocate
 *;       an entire register to the keep-alive bit, so that the state of
 *;       other bits in the register are not a concern.
 *;
 * Monitor Fields:
 *	Port:  computer output port
 *	BaudRate:  9600
 *	Parity:  0
 *	DataBits:  7
 *	StopBits:  1
 *	IDSubString:  LSCI,MODEL218 (monitor identifying substring,
 *		reply to *IDN?)
 *	PollAllCmd:  KRDG? (command to read all inputs)
 *	ChanName[1-8]:  Channel names (not used by service, just
 *		display)
 *	TempCurve[1-8]:  T curve to use (section name)
 *	PLCAddr[1-8]:  Starting memory addresses in PLC
 *	CfgStrCount:  Number of user strings to send to monitor.
 *	CfgStr[1-CfgStrCount):  User string to send.
 *;
 *; User configuration strings.  These are sent to the temperature
 *; monitor as part of the configuration process when the service
 *; is started.
 *;
 *;     CfgStrCount        :  number of user config strings
 *;     CfgStr1..CfgStrN   :  text of config strings to send
 *;
 *; Any number of config strings may be used, no upper limit exists.
 *; A full list of commands for a LakeShore 218 temperature monitor can
 *; be found in 218V1.6.PDF, which should be in the same directory as
 *; this file.
 *;
 *; At a minimum, CfgStr entries should be included that set the input
 *; temperature probe type for the two input banks, using the INTYPE
 *; command.
 *;
 *; The service configures each temperature monitor as follows:
 *;
 *;      1. Sets monitor to default values via "DFLT 99" command
 *;         (then waits 1.5 seconds for monitor to complete reset)
 *;      2. Sets monitor date/time to current date/time
 *;      3. "Blanks" the 8 input displays
 *;      4. Programs in temperature curves
 *;         (waits 1 second after each curve, to allow monitor to process)
 *;      5. Sets 8 input displays to show degrees K
 *;  ->  6. Sends user configuration strings
 *;      7. Assigns temperature curves to inputs
 *;
 *; Temperature curves are assigned LAST, so that INTYPE commands in the
 *; user configuration section here can properly set the input types of
 *; the A and B input banks.
 *;
 ***************************************************************/
int monitor_initialization(void) {

    int baud_rate = 0;
    int parity = 0;
    int data_bits = 0;
    int stop_bits = 0;
    char id_sub_string[kMaxLength]; 
    int cfg_str_count = 0;  /* Number of config strings according to
			       ini file */
    char** cfg_str = NULL;   /* Array to store config strings */
    int n_cfg_str = 0;       /* Number of config strings stored */
    char* temp_curve[kNumChannels];   /* Array of temp curve names
					 to use */
    int i = 0;
    for (i = 0; i < kNumChannels; i++)
	temp_curve[i] = NULL;

    /* Open initialization file for reading */
    FILE* ini_file = NULL;
    ini_file = fopen(ini_filename, "r");
    if ( !ini_file ) {
	fprintf(stderr, "Unable to open %s.\n", ini_filename);
	return(1);
    }

    /* Allocate space to read in a line */
    char *ini_line;
    int length = 100;
    ini_line = (char*) malloc(length);
    char* ini_line_save = ini_line;
    int size = 10000;

    int ret = find_section(ini_file, monitor_id);
    if (ret != 0) {
	exit(1);
    }

    /* Loop until next section header is encountered */
    printf("Parsing input file %s for %s . . .\n", 
	   ini_filename, monitor_id);
    while ( 1 ) {
		
	ini_line = ini_line_save;
	size = getline(ini_line, length, ini_file);
	if ( feof(ini_file) )
	    break;
		
	/* Check for section header */
	if (ini_line[0] == kSectionOpen)
	    break;
		
	/* Skip comment lines (begin with ;) or blank lines*/
	if (ini_line[0] == kCommentMarker || !isalpha(ini_line[0])) {
	    continue;
	}
		
	/* Populate TempMonitor_t struct with appropriate
	   information */
	char* char_ptr;
	if ((char_ptr = key_value(ini_line, "Port"))) {
	    continue;
	} else if ((char_ptr = key_value(ini_line, "BaudRate"))) {
	    baud_rate = atoi(char_ptr);
	} else if ((char_ptr = key_value(ini_line, "Parity"))) {
	    parity = atoi(char_ptr);
	} else if ((char_ptr = key_value(ini_line, "DataBits"))) {
	    data_bits = atoi(char_ptr);
	} else if ((char_ptr = key_value(ini_line, "StopBits"))) {
	    stop_bits = atoi(char_ptr);
	} else if ((char_ptr = key_value(ini_line, "IDSubString"))) {
	    strcpy(id_sub_string, char_ptr);
	} else if ((char_ptr = key_value(ini_line, "PollAllCmd"))) {
	    strcpy(poll_all_cmd, char_ptr);
	} else if ((char_ptr = key_value(ini_line, "CfgStrCount"))) {
	    cfg_str_count = atoi(char_ptr);
	    cfg_str = (char**) malloc(cfg_str_count*sizeof(char*));
	    if ( cfg_str == 0) {
		printf("No more room in the inn.\n");
	    }
	    int i = 0;
	    for (i = 0; i<cfg_str_count; i++)
		cfg_str[i] = NULL;
	} else if ((char_ptr = key_value(ini_line, "CfgStr"))) {
	    /* Order is important.  Placing this case after
	       CfgStrCount means I only have to deal with the
	       strings.  Count has already been read in. */

	    if (cfg_str_count == 0) {
		fprintf(stderr, "Initialization Error!  CfgStr strings present!\n");
		return(1);
	    }

	    if (n_cfg_str >= cfg_str_count) {
		fprintf(stderr, "Warning:  # of configuration strings >= CfgStrCount.  Ignoring remaining strings.\n");
		continue;
	    }
		    
	    char* ret = NULL;
	    ret = malloc(strlen(char_ptr)+1);
	    strcpy(ret, char_ptr);
	    cfg_str[n_cfg_str] = ret;
	    n_cfg_str++;
	} else if ((char_ptr = key_value(ini_line, "ChanName"))) {
	    continue;
	} else if ((char_ptr = key_value(ini_line, "TempCurve"))) {
	    int i;
	    ini_line += strlen("TempCurve");
	    i = strtol(ini_line, NULL, 0);
	    if (i > kNumChannels) {
		fprintf(stderr, "Exceeded maximum allowable number of channels.  Exiting monitor_initialization . . .\n");
		return(1);
	    }
	    temp_curve[i-1] = (char*)malloc(strlen(char_ptr)+1);
	    strcpy(temp_curve[i-1], char_ptr);
	} else if ((char_ptr = key_value(ini_line, "PLCAddr"))) {
	    int i;
	    ini_line += strlen("PLCAddr");
	    i = strtol(ini_line, NULL, 0);
	    if (i > kNumChannels) {
		fprintf(stderr, "Exceeded maximum allowable number of channels.  Exiting monitor_initialization . . .\n");
		return(1);
	    }
	    plc_addr[i-1] = strtol(char_ptr, NULL, 0);
	} else {
	    /* Should never get here */
	    fprintf(stderr, "Should never have gotten here!\n");
	    fprintf(stderr, "ini_line = %s\n", ini_line);
	    exit(1);
	}
    } /* End loop over section */

    /* Can establish communications with temperature monitor
       now */
    fd_serial = open_port(serial_port);
    if (fd_serial < 0) {
	fprintf(stderr, "monitor_initialization:  Failed to open serial port.  Exiting . . .\n");
	exit(1);
    }

    int i_out = 0;
    char s_text[length];

    /* Reset monitor (do not erase temperature curves) */
    i_out = write_port(fd_serial, "DFLT 99");
    sleep(2);

    /* Set date and time */
    time_t curtime;
    struct tm *loctime;
    curtime = time(NULL);
    loctime = localtime(&curtime);
    sprintf(s_text, "DATETIME %d, %d, %d, %d, %d, %d", 
	    loctime->tm_mon+1, loctime->tm_mday, 
	    loctime->tm_year-100, 
	    loctime->tm_hour, loctime->tm_min, loctime->tm_sec);
    i_out = write_port(fd_serial, s_text);

    /* Blank display */
    for (i = 0; i<kNumChannels; i++) {
	sprintf(s_text, "DISPFLD %d, 0", i+1);
	i_out = write_port(fd_serial, s_text);
    }

    /* Program in temperature curves */
    for (i=0; i<kNumChannels; i++) {  /* Loop over all channels */

	printf("Programming monitor %d curve %d:  %s . . .\n", 
	       monitor_num, i+1, temp_curve[i]);

	/* Got to beginning of ini file */
	rewind(ini_file);

	/* Search for section */
	int ret = find_section(ini_file, temp_curve[i]);
	if (ret != 0) {
	    exit(1);
	}

	/* Loop over file gathering temperature curve data
	   points until encounter EOF or another section
	   header */
	int n_curve_points = 0; /* Actual number of points */
	char serial_no[kMaxLength];  /* Curve serial number */
	strncpy(serial_no, "\0", kMaxLength);
	int format = 0;         /* Curve data format:
				   2 = V/K, 3 = Ohm/K, 
				   4 = log Ohm/K */
	int limit_kelvin = 0;   /* Curve limit in K */
	int coefficient = 0;    /* Curve T coefficient:
				   1 = negative
				   2 = positive */
	int point_count = 0;    /* Number of points according
				   to ini file */
	char point[kMaxLength]; /* Curve data point */
	strncpy(serial_no, "\0", kMaxLength);
	int n_point = 0;        /* Number of point */
	char* char_ptr;
		
	while (1) { /* Loop over temperature curve points */

		    /* Read in line of file */
	    ini_line = ini_line_save;
	    size = getline(ini_line, length, ini_file);
		    
	    /* Check for end of file or section header */
	    if ( feof(ini_file) || ini_line[0]==kSectionOpen ) {
		/* Check to make sure some points were read in
		 */
		if ( n_curve_points > 0 ) {
		    break;
		} else {
		    fprintf(stderr, "**monitor_initialization => Unable to find temperature curve data points.  Exiting . . .\n");
		    exit(1);
		}
	    } /* End EOF or kSectionOpen check */

	    /* Check for valid entries and parse */
	    if ((char_ptr = key_value(ini_line, "SerialNo"))) {
		strcpy(serial_no, char_ptr);
	    } else if ((char_ptr = key_value(ini_line, "Format"))) {
		format = atoi(char_ptr);
	    } else if ((char_ptr = 
			key_value(ini_line, "LimitKelvin"))) {
		limit_kelvin = atoi(char_ptr);
	    } else if ((char_ptr = 
			key_value(ini_line, "Coefficient"))) {
		coefficient = atoi(char_ptr);
	    } else if ((char_ptr = 
			key_value(ini_line, "PointCount"))) {
		point_count = atoi(char_ptr);

		/* Clear existing user curve */
		sprintf(s_text, "CRVDEL %d", 20+i+1);
		i_out = write_port(fd_serial, s_text);

		/* Send curve header info to serial port */
		sprintf(s_text, "CRVHDR %d,%s,%s,%d,%d,%d",
			20+i+1, temp_curve[i], serial_no, format, 
			limit_kelvin, coefficient);
		//		printf("**monitor_initialization => %s\n", s_text);
		i_out = write_port(fd_serial, s_text);

	    } else if ((char_ptr = 
			key_value(ini_line, "Point"))) {
		strcpy(point, char_ptr);
		n_curve_points++;

		/* Extract point number */
		ini_line += strlen("Point");
		n_point = atoi(ini_line);

		/* Send curve point to serial port */
		sprintf(s_text, "CRVPT %d,%d,%s", 20+i+1,
			n_point, point);
		i_out = write_port(fd_serial, s_text);
			
	    } else {
		;
	    }

	} /* End temperature curve loop */

	sleep(1);
		    
    } /* End loop over channels */

    /* Set display to show readings in Kelvin */
    for (i = 0; i<kNumChannels; i++) {
	sprintf(s_text, "DISPFLD %d, %d, 1", i+1, i+1);
	i_out = write_port(fd_serial, s_text);
    }

    /* Send user configuration strings */
    for (i = 0; i<n_cfg_str; i++) {
	i_out = write_port(fd_serial, cfg_str[i]);
    }

    /* Assign temperature curves to inputs */
    for (i = 0; i<kNumChannels; i++) {
	sprintf(s_text, "INCRV %d,%d", i+1, 20+i+1);
	i_out = write_port(fd_serial, s_text);
    }
	    
    /* Check to make sure output pipe already exists and is writeable
       */
    int err = access(pipe_filename, F_OK);
    if (err != 0) {
	perror("monitor_initialization()");
	return -1;
    }

    err = access(pipe_filename, W_OK);
    if (err != 0) {
	perror("monitor_initialization()");
	return -1;
    }
    printf("Monitor %d:  File %s exists and is available for writing . . .\n", 
	   monitor_num, pipe_filename);
    
    /* Open output file */
    out_file = fopen(pipe_filename, "w");
    if (out_file == NULL) {
	fprintf(stderr, 
		"**monitor_initialization => Unable to open %s . . .\n",
		pipe_filename);
	return -1;
    }
    printf("Monitor %d:  Successfully opened %s for writing . . .\n", 
	   monitor_num, pipe_filename);

    /* Set line buffering on out_file stream */
    err = setvbuf(out_file, NULL, _IOLBF, BUFSIZ);
    if (err != 0) {
	perror("monitor_initialization()");
	return -1;
    }

    /* Free up space */
    free(ini_line); ini_line = NULL;

    for (i = 0; i < n_cfg_str; i++) {
	free(cfg_str[i]); cfg_str[i] = NULL;
    }
    
    free(cfg_str); cfg_str = NULL;
    
    for (i = 0; i < kNumChannels; i++) {
	free(temp_curve[i]); temp_curve[i] = NULL;
    }
    
    fclose(ini_file);
    
    //    printf("Leaving monitor_initialization()\n");
    return(0);
}

/****************************************************************
 *
 **************************************************************/
int monitor_poll(void) {

    /* Declaration and Initialization */
    int length = 100;
    char s_text[length];
    strncpy(s_text, "\0", length);
    double temp[kNumChannels];
    int i = 0;
    
    /* Query monitor and get response */
    int ret = write_port(fd_serial, poll_all_cmd);
    ret = read_port(fd_serial, s_text, length);

    /* Process return string */
    char* s_ret = NULL;
    s_ret = strtok(s_text, ",");
    while (s_ret != NULL) {
	temp[i] = atof(s_ret);
	i++;
	s_ret = strtok(NULL, ",");
    }

    /* Check to make sure all channels were read back.  Return with
       error if not */
    if (i != kNumChannels) {
	printf("**monitor_poll => Only read back %d channels . . .\n", i);
	return -1;
    }

    /* Render temperature data in IEEE 4 Byte Real format */
    
    /* Write data into appropriate file */
    //    printf("**********************\n");
    //    rewind(out_file);  /* Rewind output file to beginning */
    for (i = 0; i < kNumChannels; i++) {
	//	fprintf(stdout, "**monitor_poll => %d %d %d %f\n", 
	//		monitor_num, i+1, plc_addr[i], temp[i]);
	int err = fprintf(out_file, "%d %d %d %f\n", monitor_num, 
			  i+1, plc_addr[i], temp[i]);
	/* Check for writing errors and bomb out */
	//	printf("err = %d\n", err);
	if (err < 0) {
	    perror("monitor_poll()");
	    return -1;
	}
    }

    return 0;
}

///***************************************************************
// * This function reads a line of input from file and returns the
// * length.  Chops off the newline character.
// **************************************************************/
//int getline(char *line, int max, FILE* file) {
//if (fgets(line, max, file) == NULL)
//     return 0;
//     else {
////	fprintf(stderr, "getline %d:  %s", strlen(line), line);
//char* char_ptr;
//char_ptr = strpbrk(line, "\r\n");
//*char_ptr = '\0';
////	fprintf(stderr, "getline %d:  %s\n", strlen(line), line);
//return strlen(line);
//}
//}
//
///***************************************************************
// * This function returns a pointer to the value of a key-value pair
// * separated by an equals sign and one or more spaces iff key matches
// * second argument.
// * Return = NULL if not a match
// * Return = pointer string if match found
// **************************************************************/
//char* key_value(char* ini_line, const char* key) {
//    
//    char* char_ptr;
//    if (strncmp(ini_line, key, strlen(key)) == 0) {
//	char_ptr = strpbrk(ini_line, "=");
//	char_ptr++;
//	if (!isalnum(*char_ptr)) {
//	    char_ptr++;
//	}
//	return char_ptr;
//    } else {
//	return NULL;
//    }
//}
//	
///***************************************************************
// * This function positions the file pointer on the line containing the
// * specified section header.
// * Return = 0 if section header founder
// * Return = -1 if reach EOF without finding section header
// **************************************************************/
//int find_section(FILE* file, const char* section) {
//
//    int length = 100;
//    char ini_line[length];
//    char section_head[strlen(section)+3];
//    int size = 0;
//
//    sprintf(section_head, "%c%s%c", kSectionOpen, section, kSectionClose);
//
//    while (1) { /* Loop for section header */
//	
//	/* Read in line of file */
//	size = getline(ini_line, length, file);
//	
//	/* Check for end of file */
//	if ( feof(file) ) {
//	    fprintf(stderr, "**find_section => Unable to find %s.\n", 
//		    section_head);
//	    return -1;
//	}
//	
//	/* Check to see if the line contains the section
//	   header */
//	if (strncmp(ini_line, section_head, 
//		    strlen(section_head)) == 0) {
//	    return 0;
//	}
//	
//    } /* End section header loop */
//
//}
